import model
import optimization
import json
import numpy as np
import os
import re
import postpro
import matplotlib.pyplot as plt
import plotting
from datetime import datetime

# returns a sorted list of all optimization log files
def list_optimization_results(log_path = "logs/", date_time_regex = "\d\d\d\d-\d\d-\d\d_\d\d-\d\d-\d\d"):
	regex = "optimization_log__" + date_time_regex + "\.json"

	l = [log_path+f for f in os.listdir(log_path) if re.search(regex, f)]
	l.sort()

	return l


# auto completes fields from older versions
def json_import(fn):
	d = json.load(open(fn, "r"))

	#those have to be in a log file
	assert("start_time" in d)
	assert("initialization" in d)
	if "result" not in d:
		raise Exception("incomplete log file, optimization did not finish:", fn)

	if "parameterization" in d: # old type:
		d["parametrization"] = d.pop("parameterization")

	# nothing to complete for now...
	if "meta_info" not in d:
		d["meta_info"] = {"scene_file" : "not_specified"}

	return d


def load_log_setups(fn):
	# old format only stored just a combined camera & laser position, so we might need to fix this
	d = json_import(fn)

	assert("meta_info" in d)
	assert("scene_file" in d["meta_info"])
	assert(os.path.isfile(d["meta_info"]["scene_file"]))

	init = model.setup_from_json_dict(d["initialization"])
	result = model.setup_from_json_dict(d["result"])
	gt = model.load_setup_3d_from_file(d["meta_info"]["scene_file"])

	return gt, init, result


def get_noise_from_log(log_fn):
	d = json_import(log_fn)
	info = d["meta_info"]
	if "init_noise" in info:
		return d["meta_info"]["init_noise"], d["meta_info"]["tof_noise"]
	else:
		return d["meta_info"]["noise"], d["meta_info"]["tof_noise"]


def compute_rms(gt, result):
	g_S, g_L, g_C, _, _ = gt
	r_S, r_L, r_C, _, _ = result

	G = np.vstack((g_S, g_L, g_C))
	R = np.vstack((r_S, r_L, r_C))

	diff = np.linalg.norm(G-R, axis=1)
	rms = np.sqrt(np.sum(diff**2) / len(diff))

	return rms


def compute_minimized_rms(gt, result):
	opt = postpro.least_squares_transform(gt, result)
	return compute_rms(gt, opt)

def compute_minimized_rms_from_log(log_fn):
	gt, init, res = load_log_setups(log_fn)
	return compute_minimized_rms(gt, res)

#def compute_rms_from_log(log_fn):
#	gt, init, res = load_log_setups(log_fn)
#
#	return compute_rms(gt, res)

def get_optimization_time_seconds(start_string, end_string):
	time_fmt = "%Y-%m-%dT%H:%M:%S"
	time = datetime.strptime(end_string, time_fmt) - datetime.strptime(start_string, time_fmt)
	time = time.seconds
	return time



def evaluate_reconstruction(log_fn=None, show_mirrors=True):
	if not log_fn:
		log_fn = list_optimization_results(log_path="logs/tof_noise_0-0.1/")[9]
		print(log_fn)

	gt, init, res = load_log_setups(log_fn)

	# there is a case, when not all lasers or mirrors from the gt are used. We accommodate for that here:
	if gt.L.shape != init.L.shape or gt.M.shape != init.M.shape:
		model.set_number_of_lasers_and_mirrors(gt, len(init.L), len(init.M))

	rms = compute_minimized_rms(gt, res)
	transf = postpro.least_squares_transform(gt, res)

	noise, tof_noise = get_noise_from_log(log_fn)
	print("initialization noise:", noise)
	print("tof noise:", tof_noise)
	print("RMS:", rms)

	plotting.draw_setup_3d(gt)
	plt.title("gt")

	plotting.draw_setup_3d(init, show_mirrors=show_mirrors)
	plt.title("initialization")

	plotting.draw_setup_3d(res, show_mirrors=show_mirrors)
	plt.title("reconstruction")

	plotting.draw_setup_3d(transf, show_mirrors=show_mirrors)
	plt.title("transformed reconstruction")

	plotting.compare_two_setups(gt, transf, show_ids=False, show_S=True, show_mirrors=show_mirrors)


def extract_log_key_values(path = "logs/"):
	import tabulate

	l = list_optimization_results(path)
	rows = list()
	column_names = ["i", "file name", "setup", "time", "parametrization", "input", "output", "connections", "iterations", "init_noise", "tof_noise", "mirrors", "lasers", "other"]
	for i, fn in enumerate(l):
		try:
			d = json_import(fn)
			time = get_optimization_time_seconds(d["start_time"], d["end_time"])
			scene_file = d["meta_info"]["scene_file"]
			scene_file = scene_file[scene_file.rfind("/")+1 : scene_file.rfind(".")]
			parametrization = d["parametrization"]["type"]
			input_d = d["parametrization"]["input_dimensions"]
			output_d = d["parametrization"]["output_dimensions"]
			if "connection_idx" in d["meta_info"]:
				connections = len(d["meta_info"]["connection_idx"]) // np.sum(d["meta_info"]["connection_idx"])
				con_str = "1/{}".format(connections)
			else:
				con_str = "full"
			iterations = d["optimizer"]["iterations"]
			if "init_noise" in d["meta_info"]:
				init_noise = d["meta_info"]["init_noise"]
			else:
				init_noise = d["meta_info"]["noise"]
			tof_noise = d["meta_info"]["tof_noise"]
			mirrors = len(d["initialization"]["mirrors"])
			lasers = len(d["initialization"]["laser_points"])

			other = ""
			# can be used to read additional information from the meta_info dict

			#rms = compute_rms_from_log(fn)
			rows.append((i, fn[fn.rfind("_20"):], scene_file, time, parametrization, input_d, output_d, con_str, iterations, init_noise, tof_noise, mirrors, lasers, other))
		except Exception as e:
			print("error parsing", fn)
			print(e)
	#print(tabulate.tabulate(rows, headers=column_names))
	print(tabulate.tabulate(rows, headers=column_names, tablefmt="github"))
	print("Elements:", len(rows))

	return rows
