import numpy as np
import matplotlib.pyplot as plt
from math_util import *
import plotting
import model
import optimization
import evaluation
import os


def calibrate_example_setup():
	# Overview:
	# This function loads a setup, disturbs it by applying noise and then attemps to recreate the original
	# setup by applying our calibration scheme.

	scene_file = "scenes/scene.json"
	setup = model.load_setup_3d_from_file(scene_file)

	# The setup contains 8 lasers and 40 mirrors, which takes quite some time to calibrate
	# Thus, the setup is reduced to more common dimensions.
	num_lasers = 5
	num_mirrors = 6
	model.set_number_of_lasers_and_mirrors(setup, num_lasers, num_mirrors)

	# this function can be used to visualize setups (ground truth, calibrations, etc.)
	plotting.draw_setup_3d(setup, show_ids=False, show_mirrors=True)
	plt.show()

	# create time-of-flight measurementes.
	# For a real calibration, these would be the camera measurements
	tof_noise = 0.05
	tof_ref = model.generate_measurements(setup)
	tof_ref = np.random.normal(loc=tof_ref, scale=tof_noise)

	# create an initialization from the ground truth setup
	# For a real calibration, the initialization would be estimated by the user.
	init_noise = 0.2
	setup_n = model.apply_noise_to_setup(setup, noise=init_noise)

	# start the optimization. The output is written to the "logs/" directory
	# log_meta_info contains information such as noise levels and the
	# ground truth scene file. These will be missing for in real calibrations,
	# but other information that might be useful later on can also be stored here.
	os.makedirs("logs/", exist_ok=True)
	opt = optimization.joint_optimization(
		setup_n, tof_ref,
		parametrization="planar",
		log_meta_info={"init_noise": init_noise, "tof_noise": tof_noise, "scene_file": scene_file})
	print("done")



def calibrate_example_grid():
	# when the grid parametrization is used, the sensor pattern must be provided to the optimize

	init_noise = 0.2
	# get_sensor_pattern_for_setup can return the ground truth pattern (noise=0)
	# or a version with noise applied to it
	pattern_gt = model.get_sensor_pattern_for_setup("scene.json", 0)
	pattern_noise = model.get_sensor_pattern_for_setup("scene.json", init_noise)

	# preview the pattern and the initialization
	plt.figure()
	plt.scatter(pattern_gt[:,0], pattern_gt[:,1])
	plt.scatter(pattern_noise[:,0], pattern_noise[:,1])
	plt.show()

	# this code is mostly identical to 'calibrate_example_setup' (see above)
	scene_file = "scenes/scene.json"
	setup = model.load_setup_3d_from_file(scene_file)
	num_lasers = 5
	num_mirrors = 6
	model.set_number_of_lasers_and_mirrors(setup, num_lasers, num_mirrors)
	tof_noise = 0.05
	tof_ref = model.generate_measurements(setup)
	tof_ref = np.random.normal(loc=tof_ref, scale=tof_noise)

	setup_n = model.apply_noise_to_setup(setup, noise=init_noise)

	os.makedirs("logs/", exist_ok=True)
	opt = optimization.joint_optimization(
		setup_n, tof_ref,
		parametrization="grid", # enable grid parametrization
		c_pattern = pattern_noise, # set the sensor pattern
		log_meta_info={"init_noise": init_noise, "tof_noise": tof_noise, "scene_file": scene_file})
	print("done")


def plot_calibration_result():
	# this convenience function returns a list of all log files in a directory
	logs = evaluation.list_optimization_results(log_path="logs/")
	if 0 == len(logs):
		raise Exception("no log files found")

	# if multiple logs are in a folder, this function is useful to give a quick overview of each
	evaluation.extract_log_key_values("logs/")
	print("\n\n")

	# this convenience function compares a calibration to the ground truth value
	"""
	The ground truth setup is read from the log_meta_info dict (see above)
	5 Plots are created:
	- the ground truth setup
	- the initialization of the optimisation
	- the calibrated setup (with is potentially rotated)
	- the calibration result transformed to match the ground truth as closely as possible
	- direct comparison between transformed calibration and ground truth
	Additionally some information like the calibration RMS is printed.
	
	Since ground truth information is not available for real calibrations, this function cannot
	be used in such cases. Instead, the calibrated setup could be displayed via
	`plotting.draw_setup_3d` (see above).
	"""
	evaluation.evaluate_reconstruction(logs[0])


if __name__ == "__main__":
	calibrate_example_setup()
	calibrate_example_grid()
	plot_calibration_result()
	plt.show()
