import autograd.numpy as np
import numpy.random
from math_util import *
import json
from recordclass import recordclass, RecordClass #pip install recordclass

# S: position of laser (S[0]) and camera (S[1])
# L: a laser point
# C: some camera points.
# M: reflection plane
# connection_list: [(L, C, M), ...] pairs of indices
Setup = recordclass("Setup", ['S', 'L', 'C', 'M', 'connection_list'], defaults=(None,)*5)


# function that computes the travel time:
def f(S, l, c, m):
	l_prime = reflect_point_on_plane(l, m)

	dist = (
			np.linalg.norm(l-S[0]) +
			np.linalg.norm(c-l_prime) +
			np.linalg.norm(S[1]-c)
	)
	return dist


# just like f, but does not consider the travel distance to the setup
def g(S, l, c, m):
	l_prime = reflect_point_on_plane(l, m)
	dist = np.linalg.norm(c-l_prime)
	return dist


def generate_measurements(setup):
	t = list() # list, as autograd does not support array assignment
	M = [normalize_plane(m) for m in setup.M] # make a copy of normalized planes
	for connection in setup.connection_list:
		l, c, m = connection
		t.append(f(setup.S, setup.L[l], setup.C[c], M[m]))
	return np.array(t)


def create_fully_connected_list(setup):
	# ignores setup.connection_list and returns a new one, using all cameras, lasers and mirrors
	connection_list = np.empty((setup.C.shape[0] * setup.L.shape[0] * setup.M.shape[0], 3), dtype=int)
	idx = 0
	for l in range(setup.L.shape[0]):
		for c in range(setup.C.shape[0]):
			for m in range(setup.M.shape[0]):
				connection_list[idx, :] = l, c, m
				idx += 1
	return connection_list


# d: a json dict loaded from a file
def setup_from_json_dict(d):
	setup = Setup()
	setup.C = np.array(d["camera_points"], dtype=np.float)
	setup.L = np.array(d["laser_points"], dtype=np.float)
	setup.M = np.array(d["mirrors"], dtype=np.float)
	setup.S = np.array(d["setup"], dtype=np.float)

	# check if file stored camera and laser at the same position
	if setup.S.shape == (3,):
		setup.S = np.array([setup.S, setup.S])

	# check if the file stored connection information
	if "connection_list" in d:
		setup.connection_list = np.array(d["connection_list"], dtype=np.int)
	else: # create fully connected list
		setup.connection_list = create_fully_connected_list(setup)
	return setup


def load_setup_3d_from_file(fn):
	assert(fn[-5:] == ".json")
	f = json.load(open(fn, "r"))
	return setup_from_json_dict(f)


def set_number_of_lasers_and_mirrors(setup, l, m):
	assert(len(setup.connection_list) == len(setup.L)*len(setup.C)*len(setup.M)) # only for fully connected setups
	setup.L = setup.L[:l]
	setup.M = setup.M[:m]
	setup.connection_list = create_fully_connected_list(setup)  # recreate connections:


def save_setup(fn, setup, meta_info=None):
	assert(fn[-5:] == ".json")
	d = {
		"camera_points": setup.C.tolist(),
		"laser_points": setup.L.tolist(),
		"mirrors": setup.M.tolist(),
		"setup": setup.S.tolist()
	}
	if meta_info is not None:
		d["meta_info"] = meta_info
	with open(fn, "wt") as f:
		json.dump(d, f, indent="\t")

################################################################


def apply_noise_to_setup(setup, noise):
	#S_n = np.random.normal(S, noise)
	setup_noise = Setup()
	setup_noise.S = setup.S
	setup_noise.L = np.random.normal(setup.L, noise)
	setup_noise.C = np.random.normal(setup.C, noise)

	# the mirrors are randomly rotated around the center of the setup
	center = np.average(np.vstack((setup.L, setup.C)), axis=0)

	setup_noise.M = np.empty(setup.M.shape)
	for i, m in enumerate(setup.M):
		# rotate plane around center
		old_dist = distance_to_plane(center, m)

		n = m[:-1]
		# since n is a direction rather than a point, it is same to define a 'same level of noise' but for now this should do
		new_n = np.random.normal(n, noise/4)
		new_n = new_n / np.linalg.norm(new_n)

		# restore distance to center
		new_dist = np.dot(center, new_n)-old_dist
		new_dist = np.random.normal(new_dist, noise)

		setup_noise.M[i]=([*new_n, -new_dist])

	setup_noise.connection_list = setup.connection_list.copy() # not really important to copy here

	return setup_noise


################################################

# renormalizes all planes
def normalize_mirror_planes(setup):
	for i, m in enumerate(setup.M):
		#m = normalize_plane(m)
		setup.M[i] = normalize_plane(m)

# e.g. reduce = 5: returns 1/5th of the connections
# seed = None : random seed
def remove_random_connections(setup, tof, reduce, seed=None):
	rnd = numpy.random.RandomState(seed) # important, so we don't mess with the global state here

	remain = len(setup.connection_list) // reduce
	idx = [True] * remain + [False] * (len(setup.connection_list) - remain)
	rnd.shuffle(idx)

	setup.connection_list = setup.connection_list[idx, :]
	tof = tof[idx]
	return setup, tof, idx

####################################################


import homography

def apply_noise_to_pattern(corner_points, pattern, noise):
	noisy_corners = np.random.normal(loc=corner_points, scale=noise)
	H = homography.compute_homography(corner_points, noisy_corners)
	return homography.homographic_transform(H, pattern)

import matplotlib.pyplot as plt
def get_sensor_pattern_for_setup(setup_fn, noise):
	if "scene" in setup_fn:
		X, Y = np.meshgrid(np.arange(5), np.arange(5))
		p = np.stack((X, Y))
		p = np.moveaxis(p, 0, 2).reshape((-1,2))

		# p contains normalized coordinates. now pretransform it by a homography to roughly fit the sensor pixels
		c_wall = np.array([[-1, 1], [1, 1], [1, -1], [-1, -1]])
		c_sensor= np.array([[0, 4], [4, 4], [4, 0], [0, 0]])
		c_sensor_noise = np.random.normal(loc=c_sensor, scale=noise)
		H = homography.compute_homography(c_sensor_noise, c_wall)
		pattern = homography.homographic_transform(H, p)
	else:
		raise Exception("unknown setup")

	return pattern


#############################################################


def export_setup_to_matlab(mat_name, setup, save=False, show=False, pixel_mask=None):
	import scipy.io

	if not mat_name.endswith(".mat"):
		mat_name += ".mat"

	C = setup.C
	L = setup.L
	S = setup.S


	# camera points:
	# Columns: U, V, X, Y, Z, V
	cam = np.empty((len(C), 6))
	sensor_positions = np.array(np.meshgrid(np.arange(32), np.arange(32)))[::-1].T.reshape((-1, 2))
	cam[:, (0,1)] = sensor_positions
	cam[:, (2,3,4)] = C
	if pixel_mask is not None:
		cam[:, 5] = pixel_mask
	else:
		cam[:, 5] = 1

	laser = L
	laser_names = ["1", "2", "3", "4", "5", "6", "7"]

	setup_pos = S

	if save:
		scipy.io.savemat(mat_name, {"cam": cam, "laser": laser, "laser_names": laser_names, "setup": setup_pos})
	if show:
		names = ["U Coordinate", "V Coordinate", "X Coordinate", "Y Coordinate", "Z Coordinate", "Mask"]
		for i in range(6):
			plt.figure()
			plt.title(names[i])
			plt.imshow(cam[:, i].reshape(32, 32))
			plt.colorbar()
			plt.tight_layout()
			if save:
				plot_name = mat_name[:-4]+" "+names[i]+".png"
				plt.savefig(plot_name)

