import autograd.numpy as np
from autograd import grad, jacobian

import model
import scipy.optimize
from datetime import datetime
import json
from model import Setup

def target_function(setup, t_ref):
	t_curr = model.generate_measurements(setup)
	return np.sum((t_curr-t_ref)**2)

################# scene parametrizations


### Full 3d parametrization
def pack_default(setup):
	return np.hstack((setup.L.flatten(), setup.C.flatten(), setup.M.flatten()))

def unpack_default(x, additional_params):
	l_shape, c_shape, m_shape = additional_params
	l_len = np.prod(l_shape)
	c_len = np.prod(c_shape)
	m_len = np.prod(m_shape)

	# there's got to be a nicer way for this...
	loc_L = x[ : l_len].reshape(l_shape)
	loc_C = x[l_len : l_len+c_len].reshape(c_shape)
	loc_M = x[l_len+c_len : l_len+c_len+m_len].reshape(m_shape)
	return loc_L, loc_C, loc_M



### Plane parametrization
# We asume a wall normal of 0, -1, 0 (XZ plane)
def pack_plane_parametrization(setup):
	assert(setup.L.shape[1] == 3) # plane parametrization only works in 3d
	# we assume here that the input is already kind of on the XZ plane
	p = np.vstack((setup.L, setup.C))
	avg = np.average(p, axis=0)
	dist = avg[1] # y component

	# erase y component
	# (the planes remain in the normal parametrization
	return np.hstack((
		dist,
		setup.L[:,[0,2]].flatten(),
		setup.C[:,[0,2]].flatten(),
		setup.M.flatten()))


# returns a setup derived from some parameters
def unpack_plane_parametrization(x, additional_params):
	l_shape, c_shape, m_shape = additional_params
	dist = x[0]
	l_tmp = x[1 : 1+l_shape[0]*2].reshape((l_shape[0], 2))
	c_tmp = x[1+l_shape[0]*2 : 1+l_shape[0]*2+c_shape[0]*2].reshape((c_shape[0], 2))
	M = x[1+l_shape[0]*2+c_shape[0]*2 : ].reshape(m_shape)

	# using 'repeat' looks like the nicer solution but crashes autograd :(
	#L = np.vstack((l_tmp[:,0], np.repeat(dist, l_shape[0]), l_tmp[:,1])).T
	#C = np.vstack((c_tmp[:,0], np.repeat(dist, c_shape[0]), c_tmp[:,1])).T

	L = np.vstack((l_tmp[:,0], np.array([dist]*l_shape[0]), l_tmp[:,1])).T
	C = np.vstack((c_tmp[:,0], np.array([dist]*c_shape[0]), c_tmp[:,1])).T

	return L, C, M



### Grid (on a plane) parametrization
def pack_grid_parametrization(setup):
	assert(setup.L.shape[1] == 3) # plane parametrization only works in 3d

	# this is a two step process:
	# - find the plane (as in plane parametrization)
	# - define a homography on that plane

	# plane
	p = np.vstack((setup.L, setup.C))
	avg = np.average(p, axis=0)
	dist = avg[1] # y component

	# homography (just start with identity matrix)
	H = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
	H = H.flatten()[:8] # the last element is always 1 and thus not a parameter

	# erase y component
	# (the planes remain in the normal parametrization
	return np.hstack((
		dist,
		H,
		setup.L[:,[0,2]].flatten(),
		setup.M.flatten()))



def homographic_transform(H, p):
	assert(H.shape == (3,3))
	assert(p.shape[-1] == 2)

	result = list()
	for i in range(len(p)):
		q = np.array([p[i, 0], p[i, 1], 1])
		t = np.dot(H, q)
		t /= t[2]
		result.append(t[:2])
	return np.array(result)


def unpack_grid_parametrization(x, additional_params):
	c_pattern, l_shape, m_shape, force_grid_distance = additional_params
	if force_grid_distance is not None:
		dist = force_grid_distance
	else:
		dist = x[0]
	h = x[1 : 1+8]
	l_tmp = x[1+8 : 1+8+l_shape[0]*2].reshape((l_shape[0], 2))
	M = x[1+8+l_shape[0]*2 : ].reshape(m_shape)

	# compute L
	L = np.vstack((l_tmp[:,0], np.array([dist]*l_shape[0]), l_tmp[:,1])).T

	# compute C
	H = np.hstack((h, 1)).reshape(3,3) # add the constant 1 and make it a matrix
	C_tmp = homographic_transform(H, c_pattern)
	C = np.vstack((C_tmp[:, 0], np.array([dist] * C_tmp.shape[0]), C_tmp[:, 1])).T

	return L, C, M


def compute_output_dimensions(setup, parametrization):
	if "default" == parametrization:
		return setup.L.shape[0]*3 + setup.C.shape[0]*3 + setup.M.shape[0]*4
	elif "planar" == parametrization:
		return setup.L.shape[0]*2 + setup.C.shape[0]*2 + setup.M.shape[0]*4 + 1
	elif "grid" == parametrization:
		return setup.L.shape[0]*2 + 8 + 1 + setup.M.shape[0]*4
	else:
		raise Exception("invalid parametrization: "+parametrization)


# optimizes all parameters in one big problem
# log_meta_info is forwarded to the json log
# if parametrization
def joint_optimization(setup, t_ref,
					   parametrization="default",
					   log_meta_info = None, **kwargs):
	num_input_dimensions = len(t_ref) # check dimensionality
	num_output_dimensions = compute_output_dimensions(setup, parametrization)
	if num_input_dimensions < num_output_dimensions:
		raise Exception("to few input dimensions to solve! Input {}, Output {}".format(num_input_dimensions, num_output_dimensions))
	if "c_pattern" in kwargs:
		assert(len(kwargs["c_pattern"]) == len(setup.C))

	assert(len(setup.connection_list) == len(t_ref))

	if log_meta_info is not None:
		print("log meta:", log_meta_info)

	start_time = datetime.now()
	log_fn = "logs/" + "optimization_log__" + start_time.strftime("%Y-%m-%d_%H-%M-%S")+".json"
	log = {
		"start_time" : start_time.strftime("%Y-%m-%dT%H:%M:%S")
	}
	if log_meta_info is not None:
		log["meta_info"] = log_meta_info
	log["initialization"] = {
		"setup"           : setup.S.tolist(),
		"laser_points"    : setup.L.tolist(),
		"camera_points"   : setup.C.tolist(),
		"mirrors"         : setup.M.tolist(),
		"connection_list" : setup.connection_list.tolist(),
		"tof"             : t_ref.tolist()
	}
	log["parametrization"] = {
		"type"              : parametrization,
		"input_dimensions"  : num_input_dimensions,
		"output_dimensions" : num_output_dimensions
	}
	if "grid" == parametrization:
		log["parametrization"]["c_pattern"] = kwargs["c_pattern"].tolist()


	additional_params = None
	if "default" == parametrization:
		pack = pack_default
		additional_params = (setup.L.shape, setup.C.shape, setup.M.shape)
		unpack = unpack_default
	elif "planar" == parametrization:
		pack = pack_plane_parametrization
		additional_params = (setup.L.shape, setup.C.shape, setup.M.shape)
		unpack = unpack_plane_parametrization
	elif "grid" == parametrization:
		pack = pack_grid_parametrization
		# now, the input points can't be the setup points, because those contain noise.
		# Instead they should have the same pattern as the pixels on the sensor
		c_pattern = kwargs["c_pattern"]
		if "force_grid_distance" in kwargs and kwargs["force_grid_distance"] is not None:
			additional_params = (c_pattern, setup.L.shape, setup.M.shape, kwargs["force_grid_distance"])
			log["parametrization"]["force_grid_distance"] = kwargs["force_grid_distance"]
		else:
			additional_params = (c_pattern, setup.L.shape, setup.M.shape, None)
		unpack = unpack_grid_parametrization
	else:
		raise Exception("invalid parametrization given")

	json.dump(log, open(log_fn, "w"), indent="\t")

	x_0 = pack(setup)
	x_unpack_test = unpack(x_0, additional_params)

	def fun(x, additional_params):
		# unpack variables:
		loc_L, loc_C, loc_M = unpack(x, additional_params)
		return target_function(Setup(setup.S, loc_L, loc_C, loc_M, setup.connection_list), t_ref)
	param_fun = lambda x : fun(x, additional_params)
	gradient = grad(param_fun)

	res = scipy.optimize.minimize(param_fun, x0=x_0, jac=gradient, options={"maxiter" : 500})
	L, C, M = unpack(res.x, additional_params)
	optimized_setup = Setup(S=setup.S, L=L, C=C, M=M, connection_list=setup.connection_list)


	end_time = datetime.now()
	log["end_time"] = end_time.strftime("%Y-%m-%dT%H:%M:%S")
	log["result"] = {
		"setup"           : optimized_setup.S.tolist(),
		"laser_points"    : optimized_setup.L.tolist(),
		"camera_points"   : optimized_setup.C.tolist(),
		"mirrors"         : optimized_setup.M.tolist(),
		"connection_list" : optimized_setup.connection_list.tolist(),
		"tof"             : model.generate_measurements(optimized_setup).tolist()
	}
	log["optimizer"] = {
		"success"    : res.success,
		"iterations" : res.nit,
		"message"    : res.message
	}
	json.dump(log, open(log_fn, "w"), indent="\t")

	return optimized_setup
