import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from math_util import *
import model
import optimization
import evaluation
import postpro


def plot3d_set_equal_axes_limits(points, ax):
	# since matplotlib apparently can't do equal aspect ratio for 3d plots, we fake it
	# ax.set_aspect('equal')

	size_half = np.max([
		np.max(points[:, 0]) - np.min(points[:, 0]),
		np.max(points[:, 1]) - np.min(points[:, 1]),
		np.max(points[:, 2]) - np.min(points[:, 2]),
	]
	) / 2
	x_center = (np.max(points[:, 0]) + np.min(points[:, 0])) / 2
	y_center = (np.max(points[:, 1]) + np.min(points[:, 1])) / 2
	z_center = (np.max(points[:, 2]) + np.min(points[:, 2])) / 2
	ax.set_xlim(x_center - size_half, x_center + size_half)
	ax.set_ylim(y_center - size_half, y_center + size_half)
	ax.set_zlim(z_center - size_half, z_center + size_half)

	ax.set_xlabel("x")
	ax.set_ylabel("y")
	ax.set_zlabel("z")


# computes a point on each mirror that faces at the mean of C and L
def compute_mirror_positions(setup):
	wall_center = np.mean(np.vstack((setup.C, setup.L)), axis=0)
	mirror_positions = np.empty((len(setup.M), 3))
	model.normalize_mirror_planes(setup)
	for i, m in enumerate(setup.M):
		dist = distance_to_plane(wall_center, m)
		mirror_positions[i] = wall_center - dist*m[:3]
		assert(np.abs(distance_to_plane(mirror_positions[i], m)) < 0.0001)
	return mirror_positions


def draw_setup_3d(setup, show_ids=False, show_mirrors=False):
	S, L, C, _, _ = setup

	fig = plt.figure()
	ax = fig.add_subplot(111, projection='3d')

	ax.scatter(S[0,0], S[0,1], S[0,2], color="C0", marker="<", label="setup (S)")
	ax.scatter(S[1,0], S[1,1], S[1,2], color="C0", marker="s")
	ax.scatter(L[:,0], L[:,1], L[:,2], color="C1", label="laser points (L)")
	ax.scatter(C[:,0], C[:,1], C[:,2], color="C2", label="camera points (C)")

	if show_ids:
		for i in range(L.shape[0]):
			ax.text(L[i,0], L[i,1], L[i,2], s=str(i), color="C1")
		for i in range(C.shape[0]):
			ax.text(C[i,0], C[i,1], C[i,2], s=str(i), color="C2")

	if show_mirrors:
		mirror_positions = compute_mirror_positions(setup)
		ax.scatter(mirror_positions[:,0], mirror_positions[:,1], mirror_positions[:,2], color="C3", marker="P", label="Mirror (M)")

	points = np.vstack((S, L, C))
	plot3d_set_equal_axes_limits(points, ax)


def compare_two_setups(setup1, setup2, show_S=True, show_mirrors=False, show_ids=False):
	S1, L1, C1, _, _ = setup1
	S2, L2, C2, _, _ = setup2

	fig = plt.figure()
	ax = fig.add_subplot(111, projection='3d')

	c_alpha = 0.25

	if show_S:
		ax.scatter(S1[0,0], S1[0,1], S1[0,2], color="C0", marker=">", label="setup (S)")
		ax.scatter(S1[1,0], S1[1,1], S1[1,2], color="C0", marker="s")
	ax.scatter(L1[:,0], L1[:,1], L1[:,2], color="C0", marker="^", label="laser points (L)")
	ax.scatter(C1[:,0], C1[:,1], C1[:,2], color="C0", marker="o", label="camera points (C)", alpha=c_alpha)

	if show_S:
		ax.scatter(S2[0,0], S2[0,1], S2[0,2], color="C1", marker=">", label="setup (S)")
		ax.scatter(S2[1,0], S2[1,1], S2[1,2], color="C1", marker="s")
	ax.scatter(L2[:,0], L2[:,1], L2[:,2], color="C1", marker="^", label="laser points (L)")
	ax.scatter(C2[:,0], C2[:,1], C2[:,2], color="C1", marker="o", label="camera points (C)", alpha=c_alpha)

	if show_mirrors:
		mp1 = compute_mirror_positions(setup1)
		ax.scatter(mp1[:,0], mp1[:,1], mp1[:,2], color="C0", marker="P", label="Mirror (M)")

		mp2 = compute_mirror_positions(setup2)
		ax.scatter(mp2[:,0], mp2[:,1], mp2[:,2], color="C1", marker="P", label="Mirror (M)")

	if show_ids:
		for i in range(L1.shape[0]):
			ax.text(L1[i,0], L1[i,1], L1[i,2], s=str(i), color="C0")
			ax.text(L2[i,0], L2[i,1], L2[i,2], s=str(i), color="C1")

		for i in range(C1.shape[0]):
			ax.text(C1[i,0], C1[i,1], C1[i,2], s=str(i), color="C0")
			ax.text(C2[i,0], C2[i,1], C2[i,2], s=str(i), color="C1")

	if not show_S:
		points = np.vstack((S1, L1, C1))
	else:
		points = np.vstack((L1, C1))
	plot3d_set_equal_axes_limits(points, ax)
